/* script.c
 *
 * Functions to call the DHCP client notification scripts
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "options.h"
#include "dhcpd.h"
#include "dhcpc.h"
#include "common.h"
#include "script.h"

uint8_t isValidDnsName( uint8_t* val );

/* get a rough idea of how long an option will be (rounding up...) */
static const int max_option_length[] = {
	[OPTION_IP] =		sizeof("255.255.255.255 "),
	[OPTION_IP_PAIR] =	sizeof("255.255.255.255 ") * 2,
	[OPTION_STRING] =	1,
	[OPTION_BOOLEAN] =	sizeof("yes "),
	[OPTION_U8] =		sizeof("255 "),
	[OPTION_U16] =		sizeof("65535 "),
	[OPTION_S16] =		sizeof("-32768 "),
	[OPTION_U32] =		sizeof("4294967295 "),
	[OPTION_S32] =		sizeof("-2147483684 "),
	[OPTION_RAW] =		sizeof("00 "),
	[OPTION_SIP_ENC] =	sizeof("255 "),		/* This is worst case depending on encoding */
	[OPTION_STATIC_ROUTES]= sizeof("255.255.255.255 ") * 3
};


static inline int upper_length(int length, int opt_index)
{
	return max_option_length[opt_index] *
		(length / option_lengths[opt_index]);
}


static int sprintip(char *dest, char *pre, uint8_t *ip)
{
	return sprintf(dest, "%s%d.%d.%d.%d", pre, ip[0], ip[1], ip[2], ip[3]);
}


/* really simple implementation, just count the bits */
static int mton(struct in_addr *mask)
{
	int i;
	unsigned long bits = ntohl(mask->s_addr);
	/* too bad one can't check the carry bit, etc in c bit
	 * shifting */
	for (i = 0; i < 32 && !((bits >> i) & 1); i++);
	return 32 - i;
}


/* Fill dest with the text of option 'option'. */
static void fill_options(char *dest, uint8_t *option, struct dhcp_option *type_p)
{
	int type, optlen;
	uint16_t val_u16;
	int16_t val_s16;
	uint32_t val_u32;
	int32_t val_s32;
	int len = option[OPT_LEN - 2];
	unsigned char *cp;
	int i;

	dest += sprintf(dest, "%s=", type_p->name);

	type = type_p->flags & TYPE_MASK;
	optlen = option_lengths[type];
	for(;;) {
		switch (type) {
		case OPTION_IP_PAIR:
			dest += sprintip(dest, "", option);
			*(dest++) = '/';
			option += 4;
			optlen = 4;
		case OPTION_IP:	/* Works regardless of host byte order. */
			dest += sprintip(dest, "", option);
 			break;
		case OPTION_BOOLEAN:
			dest += sprintf(dest, *option ? "yes" : "no");
			break;
		case OPTION_U8:
			dest += sprintf(dest, "%u", *option);
			break;
		case OPTION_U16:
			memcpy(&val_u16, option, 2);
			dest += sprintf(dest, "%u", ntohs(val_u16));
			break;
		case OPTION_S16:
			memcpy(&val_s16, option, 2);
			dest += sprintf(dest, "%d", ntohs(val_s16));
			break;
		case OPTION_U32:
			memcpy(&val_u32, option, 4);
			dest += sprintf(dest, "%lu", (unsigned long) ntohl(val_u32));
			break;
		case OPTION_S32:
			memcpy(&val_s32, option, 4);
			dest += sprintf(dest, "%ld", (long) ntohl(val_s32));
			break;
		case OPTION_STRING:
			memcpy(dest, option, len);
			dest[len] = '\0';
			return;	 /* Short circuit this case */
		case OPTION_SIP_ENC:
			if( type_p->code == 0x78 ){
				/* Special case SIP option 120 */
				dest = parseSipOptions( dest, option, len );
			}else{
				dest[len] = '\0';
			}
			return;	 /* Short circuit this case */
		case OPTION_RAW:
			dest += sprintf( dest, "%2.2x ", len );
			for( cp = option, i = 0; i<len; cp++,i++){
				dest += sprintf( dest, "%2.2x ", *cp );
			}
			*dest = '\0';
			return;	 /* Short circuit this case */
		case OPTION_STATIC_ROUTES:
			if(( type_p->code == 0x79 ) || ( type_p->code == 0xf9 )){
				/* Special case static routes option 121 and 249 (MS) */
				dest = parseStaticRoutesOptions( dest, option, len );
			}else{
				dest[len] = '\0';
			}
			return;	 /* Short circuit this case */
		}
		option += optlen;
		len -= optlen;
		if (len <= 0) break;
		dest += sprintf(dest, " ");
	}
}


/* put all the parameters into an environment */
static char **fill_envp(struct dhcpMessage *packet)
{
	int num_options = 0;
	int i, j;
	char **envp;
	uint8_t *temp;
	struct in_addr subnet;
	char over = 0;
	int optbufsize;

	if (packet == NULL)
		num_options = 0;
	else {
		for (i = 0; dhcp_options[i].code; i++)
			if (get_option(packet, dhcp_options[i].code)) {
				num_options++;
				if (dhcp_options[i].code == DHCP_SUBNET)
					num_options++; /* for mton */
			}
		if (packet->siaddr) num_options++;
		if ((temp = get_option(packet, DHCP_OPTION_OVER)))
			over = *temp;
		if (!(over & FILE_FIELD) && packet->file[0]) num_options++;
		if (!(over & SNAME_FIELD) && packet->sname[0]) num_options++;
	}

	envp = xcalloc(sizeof(char *), num_options + 5);
	j = 0;
	asprintf(&envp[j++], "interface=%s", client_config.interface);
	asprintf(&envp[j++], "%s=%s", "PATH",
		getenv("PATH") ? : "/bin:/usr/bin:/sbin:/usr/sbin");
	asprintf(&envp[j++], "%s=%s", "HOME", getenv("HOME") ? : "/");

	if (packet == NULL) return envp;

	envp[j] = xmalloc(sizeof("ip=255.255.255.255"));
	sprintip(envp[j++], "ip=", (uint8_t *) &packet->yiaddr);


	for (i = 0; dhcp_options[i].code; i++) {
		if (!(temp = get_option(packet, dhcp_options[i].code)))
		{
		  #ifdef IFX_DHCP_SIP
                  /* force update of sipserver to null */
                  if(dhcp_options[i].code == DHCP_SIP)
		  {
	            asprintf(&envp[j++], "sipserver=%s", "");
                  }
                  #endif
		  continue;
		}
		optbufsize = upper_length(temp[OPT_LEN - 2], dhcp_options[i].flags & TYPE_MASK) + strlen(dhcp_options[i].name) + 2;
		envp[j] = xmalloc( optbufsize );
		fill_options(envp[j++], temp, &dhcp_options[i]);

		/* Fill in a subnet bits option for things like /24 */
		if (dhcp_options[i].code == DHCP_SUBNET) {
			memcpy(&subnet, temp, 4);
			asprintf(&envp[j++], "mask=%d", mton(&subnet));
		}
	}
	if (packet->siaddr) {
		envp[j] = xmalloc(sizeof("siaddr=255.255.255.255"));
		sprintip(envp[j++], "siaddr=", (uint8_t *) &packet->siaddr);
	}
	if (!(over & FILE_FIELD) && packet->file[0]) {
		/* watch out for invalid packets */
		packet->file[sizeof(packet->file) - 1] = '\0';
		asprintf(&envp[j++], "boot_file=%s", packet->file);
	}
	if (!(over & SNAME_FIELD) && packet->sname[0]) {
		/* watch out for invalid packets */
		packet->sname[sizeof(packet->sname) - 1] = '\0';
		asprintf(&envp[j++], "sname=%s", packet->sname);
	}
	return envp;
}


/* Call a script with a par file and env vars */
void run_script(struct dhcpMessage *packet, const char *name)
{
	int pid;
	char **envp, **curr;

	if (client_config.script == NULL)
		return;

	DEBUG(LOG_INFO, "vforking and execle'ing %s", client_config.script);

	envp = fill_envp(packet);
	/* call script */
	pid = vfork();
	if (pid) {
		waitpid(pid, NULL, 0);
		for (curr = envp; *curr; curr++) free(*curr);
		free(envp);
		return;
	} else if (pid == 0) {
		/* close fd's? */

		/* exec script */
		execle(client_config.script, client_config.script,
		       name, NULL, envp);
		LOG(LOG_ERR, "script %s failed: %m", client_config.script);
		exit(1);
	}
}

uint8_t isValidDnsName( uint8_t* val )
{
	uint8_t rc = -1;

	if ( isalnum( *val ) || *val == '-' )
	{
		rc = 0;
	}

	return rc;
}


/* Parse a RFC3361 encoded buffer (src) of lenght srcLen bytes and decode it to the ascii */
/* buffer (dest) */

uint8_t *parseSipOptions( uint8_t *dest, uint8_t *src, uint16_t srcLen )
{
	uint32_t dnsTokenLen;
	uint32_t startFlag;

	/* RFC 3361 encoded list of SIP servers */
	if( *src == 0 ){
		/* Encoded as DNS name list */
		startFlag = 1;
		src++;
		srcLen--;
		while( srcLen ){
			dnsTokenLen = *src++;
			srcLen--;
			if( dnsTokenLen == 0 ){
				*dest++ = ' ';
				startFlag = 1;		/* Start of new DNS name */
			}else{
				if( startFlag == 0 ){
					*dest++ = '.';
				}
				startFlag = 0;

				if( dnsTokenLen > srcLen ){
					LOG(LOG_WARNING, "SIP option dns encoding length error\n");
					break;
				}

				while( dnsTokenLen-- ){
					// rfc1035 states the domain name can include alphanumeric or "-"
					if( isValidDnsName( src ) == 0 ){
						*dest++ = *src++;
						srcLen--;
					}else{
						LOG(LOG_WARNING, "SIP option dns char encoding error\n");
						*dest++ = '\0';
						srcLen--;
					}
				}

			}
		}
	}else{
		/* Encoded as IP address list */
		src++;
		srcLen--;
		if ((srcLen)%4 != 0 ){
			/* Error the list length should be a multiple of 4 */
			LOG(LOG_WARNING, "SIP option IP address list length not multiple of 4\n");
		}else{
			while( srcLen ){
				dest += sprintip( dest, " ", src );
				src += 4;
				srcLen -= 4;
			}
		}

	}
	*dest++ = '\0';
	return( dest );
}

/* Parse a RFC3442(Classless Static Route Option) encoded buffer
 * (src) of 0x79,scrlen, <scrLen> bytes and decode it to the ascii
 * buffer (dest) */

uint8_t *parseStaticRoutesOptions( uint8_t *result, uint8_t *src, uint16_t srcLen )
{
	unsigned int maskLen, mask, numDestBytes, dest, gw, i;
	// start
	/* Option binary format:
	 * mask [one byte, 0..32]
	 * ip [big endian, 0..4 bytes depending on mask]
	 * router [big endian, 4 bytes]
	 * may be repeated
	 *
	 * We convert it to a string "IP MASK ROUTER IP2 MASK2 ROUTER2"
	 */
	while (srcLen >= 5) { /* mask + 0-byte ip + router */

		maskLen = *src++;
        mask = dest = gw = 0;

		if (maskLen > 32)
			break;

		srcLen--;

        i = maskLen;
        while( i-- ){
            mask = (mask>>1) | 0x80000000; 
        }

        if( mask ){
		    numDestBytes =  (maskLen + 7) / 8;
            i = numDestBytes;
		    while ( i-- ){
                dest = (dest<<8) | *src++;
			    srcLen--;
            }

            if( numDestBytes < 4 )
                dest = (dest<<8);
		}

        i = 4;
        while( i-- ){
            gw = (gw<<8) | *src++;
            srcLen--;
        }

		result += sprintip(result, " ",  (uint8_t *)&dest );
		result += sprintip(result, " ", (uint8_t *)&mask );
		result += sprintip(result, " ", (uint8_t *)&gw );
	}

	*result++ = '\0';
	return( result );
}


#ifdef OPTEST

int main( int argc, char *argv[] )
{

	char dnsEncodedList[] = {0,3,'d','o','m',3,'c','o','m',0, 4, 'c','r','a','p',2,'c','o',2,'u','k',0};
	char ipEncodedList[] = {1,0,0,0,0, 137,223,234,5};

	char decodeBuf[100];

	parseSipOptions( decodeBuf, dnsEncodedList, sizeof( dnsEncodedList ) );
	printf("dnsDecode %s\n", decodeBuf );

	parseSipOptions( decodeBuf, ipEncodedList, sizeof( ipEncodedList ) );
	printf("ipDecode %s\n", decodeBuf );
}

#endif
